/*
 *	lfncnvrt: converts long filenames in makefile & corresponding directory
 *
 *	initially, designed for use with GEMLIB & WINDOM sources
 *
 *	syntax:	lfncnvrt [-u] [-f] [-nM] <makefile> <directory>
 *
 *	normal processing (-u flag NOT specified)
 *	-----------------------------------------
 *	 1.	read the specified <makefile>, saving the names of all .o
 *		files found.
 *	 2.	for each name found, check the specified <directory> for the
 *		corresponding .c or .s file.  if found, generate a new name
 *		[see below for details].
 *	 3.	create new makefile containing the new names [the old file
 *		is saved with a type of ".oldmake"]
 *	 4.	rename all the .c/.s files in <directory> to the new names.
 *	 5.	iff the -f flag is specified, generate a file in the same
 *		directory as <makefile>, containing a list of all the new
 *		.o files, one per line.  this is suitable for input to oml,
 *		the lattice librarian.
 *
 *	Note: if Fxattr() is not supported for the target <directory>,
 *	then the program will run in simulation mode: the makefile will
 *	be updated, but there will be no attempt to change names in the
 *	specified <directory>.
 *
 *	undo processing (-u flag specified)
 *	-----------------------------------
 *	The -u flag recovers the original filenames; you must specify
 *	the *current* (i.e. converted) makefile name, but the program
 *	will actually read from the original makefile (i.e. it
 *	*assumes* that the original makefile is still available with
 *	the backup extension).
 *
 *
 *	Generating a new name
 *	---------------------
 *	 1.	The maximum length of the new name stem (i.e. in front of
 *		the dot) is set by the -n option; default is 8 characters.
 *	 2.	The following transformations are applied in sequence until
 *		a valid, non-duplicate new name is generated:
 *			. convert prefix per hard-coded prefix[] rules
 *			. remove underscores
 *			. truncate
 *			. replace last char by one of the truncated chars
 *			. replace last char by '0' thru '9'
 *
 *	Version history
 *	---------------
 *	v1.0	March/2006	rfb
 *		original
 *
 *	v1.1	April/2006	rfb
 *		. add -f flag
 *		. ignore comments in makefiles, i.e. anything after a '#';
 *		  this avoids false detection of names
 *		. update prefix table to allow for new files in windom
 *
 *	v1.2	April/2006	rfb
 *		. *really* allow -f flag
 *		. update usage accordingly
 *		. only generate name in filelist if source exists
 *		. count names not found when doing undo as well as normal
 */
#include <stdio.h>
#include <stdlib.h>
#include <limits.h>
#include <mintbind.h>
#include <oserr.h>
#include <portab.h>
#include <string.h>

#define VERSION	"v1.2"

#define DEFAULT_LENGTH	8		/* of 'stem' of converted filename */
#define MAX_FILES		1000	/* that we can convert (lazy) */
#define NUM_PREFIXES	6		/* for prefix conversion (also lazy) */

#define BUFLEN			2000	/* for makefile processing */
#define NAMELEN			256

#define TOKEN_SEPARATORS	", =\\\t\n"		/* for strtok() */
#define FILLER			0xff	/* a non-zero character not used in makefiles ... */
#define COMMENT			'#'

#define NEW_EXT			"newmake"
#define BAK_EXT			"oldmake"
#define LIST_EXT		"filelist"
#define NOT_FOUND		'?'		/* extension character meaning file does not exist */

#define DOT_FREQUENCY	10

/*
 *	for generality, the following should be input from a file ...
 *	at the moment, this covers gem & windom libraries
 */
struct {
	char *old;
	char *new;
} prefix[NUM_PREFIXES] = {	"a_", "",
							"n_", "",
							"x_", "",
							"bubble_", "bub_",
							"udlib_", "udl_",
							"z_bubble_", "zbub_" };

struct convert {
	short line;					/* line number of 'origname' */
	short offset;				/* offset within line */
	char *origname;				/* without .o extension */
	char length;				/* of 'origname' */
	char ext;					/* from directory entry (.c or .s) */
	char *newname;
	unsigned short hash;		/* proxy for newname comparison */
};

struct convert lfn[MAX_FILES+1];
char buf[BUFLEN];

int filelist = FALSE;
int maxlen = DEFAULT_LENGTH;
int undo = FALSE;
int test = FALSE;				/* set if Fxattr() not supported */

int names_notfound = 0;
int names_converted = 0;
int names_smooshed = 0;
int names_truncated = 0;
int names_hinted = 0;
int names_generated = 0;
int names_unchanged = 0;

/* prototypes */
void convert_prefix(char *new,const char *original);
int create_listfile(char *filename,int n);
int create_makefile(char *filename);
unsigned short hash(const char *name);
int is_duplicate(const char *p,int n);
int load_makefile(char *filename);
int process_makefile(char *dir,int numfiles);
char *remove_uscore(char *new,char *old);
int rename_source(char *dir,int numfiles);
int replace_token(char *buf,struct convert *t);
void save_newname(struct convert *t,char *name);
int undo_changes(char *file,char *dir,int numfiles);
int update_line(char *buf,int line);
void usage(void);

int main(int argc,char **argv)
{
char file[NAMELEN], currfile[NAMELEN];
char dir[NAMELEN];
char s[NAMELEN], c;
long xattr[52/4];				/* don't need details */
int n;

	printf("lfncnvrt %s\n",VERSION);
	/*
	 *	process args
	 */
	opterr = 0;
	while((c=getopt(argc,argv,"fn:u")) != -1) {
		switch(c) {
		case 'f':
			filelist = TRUE;
			break;
		case 'n':
			maxlen = atoi(optarg);
			break;
		case 'u':
			undo = TRUE;
			break;
		case '?':
			usage();
		}
	}
	if (argc != optind+2)
		usage();

	if (undo) {
		strcpy(currfile,argv[optind]);		/* for undo, remember the current name */
		strmfe(file,currfile,BAK_EXT);		/*  but read from the backup file */
	} else strcpy(file,argv[optind]);
	strcpy(dir,argv[optind+1]);

	/*
	 *	check Fxattr() support
	 */
	sprintf(s,"%s\\x.x",dir);
	if (Fxattr(0,s,&xattr) == -EINVFN) {
		printf("Fxattr() not supported - simulation mode only\n");
		test = TRUE;
	}

	printf("Loading makefile %s\n",file);
	n = load_makefile(file);
	if (n < 0)
		exit(1);

	if (n == 0) {
		printf("No .o files found in %s\n",file);
		exit(1);
	}

	printf("%d file(s) in makefile\n",n);

	if (process_makefile(dir,n) < 0)
		exit(1);

	if (undo) {
		if (undo_changes(currfile,dir,n) < 0)
			exit(1);
	} else {
		/*
		 *	create new makefile & rename source files
		 */
		if (create_makefile(file) < 0)
			exit(1);
		if (rename_source(dir,n) < 0)
			exit(1);
		if (filelist)
			if (create_listfile(file,n) < 0)
				exit(1);
	}

	printf("%4d names not found (& thus unchanged)\n",names_notfound);
	printf("%4d names converted\n",names_converted);
	printf("%4d names smooshed\n",names_smooshed);
	printf("%4d names truncated\n",names_truncated);
	printf("%4d names hinted\n",names_hinted);
	printf("%4d names generated\n",names_generated);
	printf("%4d names could not be changed\n",names_unchanged);

	return 0;
}

void convert_prefix(char *new,const char *original)
{
const char *p;
char *q;
int i;

	for (i = 0; i < NUM_PREFIXES; i++) {
		for (p = original, q = prefix[i].old; *q; p++, q++)
			if (*p != *q)
				break;
		if (*q)				/* no match */
			continue;
		/*
		 *	match: copy new prefix followed by old body
		 */
		for (q = prefix[i].new; *q; new++, q++)
			*new = *q;
		while(*p)
			*new++ = *p++;
		*new = '\0';
		return;
	}

	strcpy(new,original);
}

int is_duplicate(const char *p,int n)
{
struct convert *t;
int i;
unsigned short h = hash(p);

	for (i = 0, t = lfn; i < n; i++, t++) {
		if (h != t->hash)
			continue;
		if (strcmp(p,t->newname) != 0)
			continue;
		return TRUE;
	}

	return FALSE;
}

void save_newname(struct convert *t,char *name)
{
	t->newname = strdup(name);
	t->hash = hash(name);
}

/*
 * hash: transform a character string into a short integer
 */
unsigned short hash(const char *name)
{
char c;
unsigned short value;

	/*
	 * take the sum of each pair of characters, and set a bit
	 * for this sum modulo the mask size (word size) in bits.
	 *
	 * e.g.	string:	c0, c1, c2, c3
	 *	set bit for (c0+c1) % mask_size
	 *	set bit for (c1+c2) % mask_size
	 *	set bit for (c2+c3) % mask_size
	 *	set bit for (c3)    % mask_size	(end condition)
	 *
	 * this gives a fair distribution, with not TOO much work
	 */
	value = 0;
	while (*name) {
		c = *name++;
		value |= 1L << ((c + *name) % (sizeof(value) * 8));
	}

	return value;
}

void usage()
{
	printf("usage: lfncnvrt [-f] [-nN] [-u] <makefile> <folder>\n");
	exit(1);
}

int load_makefile(char *filename)
{
FILE *fp;
struct convert *t;
char *token;
int line, numfiles, n;

	fp = fopen(filename,"rt");
	if (!fp) {
		printf("can't open %s\n",filename);
		return -1;
	}

	for (line = 0, numfiles = 0, t = lfn; fgets(buf,BUFLEN,fp); line++) {
		token = strtok(buf,TOKEN_SEPARATORS);
		while(token) {
			if (*token == COMMENT)
				break;
			n = strlen(token);
			if (n >= 3) {
				if (strcmp(token+n-2,".o") == 0) {
					*(token+n-2) = '\0';
					if (numfiles >= MAX_FILES) {
						printf("Too many input .o files\n");
						fclose(fp);
						return -1;
					}
					t->line = line;
					t->offset = token - buf;
					t->origname = strdup(token);
					t->length = n - 2;
					t->ext = NOT_FOUND;
					t->newname = NULL;
					t->hash = 0;
					t++;
					numfiles++;
				}
			}
			token = strtok(NULL,TOKEN_SEPARATORS);
		}
	}
	/*
	 *	for simpler processing in update_line(),
	 *	we create a dummy last_entry_plus_one
	 */
	t->line = SHRT_MAX;
	t->offset = 0;
	t->origname = "";
	t->length = 0;
	t->ext = 0;
	t->newname = "";
	t->hash = 0;

	fclose(fp);

	return numfiles;
}

int process_makefile(char *dir,int numfiles)
{
long xattr[52/4];				/* don't need details */
char s[NAMELEN], convname[NAMELEN], tempname[NAMELEN], c;
char *p, *convptr, *subpos;
struct convert *t;
int i, templen;

	if (!undo && !test)
		printf("Checking old names & generating new names ");
	else printf("Generating new names ");

	for (i = 0, t = lfn; i < numfiles; i++, t++) {
		/*
		 *	if not undoing, check now that corresponding file exists
		 */
		if (!undo && !test) {
			if (i%DOT_FREQUENCY == 0)
				printf(".");
			t->ext = 'c';
			sprintf(s,"%s\\%s.%c",dir,t->origname,t->ext);
			if (Fxattr(0,s,&xattr) < 0) {
				t->ext = 's';
				sprintf(s,"%s\\%s.%c",dir,t->origname,t->ext);
				if (Fxattr(0,s,&xattr) < 0) {
					printf("\nWarning: %s.o does not have a corresponding source file\n",t->origname);
					save_newname(t,t->origname);
					t->ext = NOT_FOUND;		/* source file does not exist */
					names_notfound++;
					continue;
				}
			}
		}

		/*
		 *	perform prefix conversion
		 */
		convert_prefix(convname,t->origname);
		if (strlen(convname) <= maxlen) {
			if (!is_duplicate(convname,i)) {
				save_newname(t,convname);
				names_converted++;
				continue;
			}
		}

		/*
		 *	remove underscores
		 */
		convptr = remove_uscore(tempname,convname);
		if (strlen(tempname) <= maxlen) {
			if (!is_duplicate(tempname,i)) {
				save_newname(t,tempname);
				names_smooshed++;
				continue;
			}
		}

		/*
		 *	truncate
		 */
		tempname[maxlen] = '\0';
		if (!is_duplicate(tempname,i)) {
			save_newname(t,tempname);
			names_truncated++;
			continue;
		}

		/*
		 *	determine substitution position for new name
		 */
		templen = strlen(tempname);
		if (templen < maxlen) {
			tempname[templen++] = '?';
			tempname[templen] = '\0';
		}
		subpos = tempname + templen - 1;

		/*
		 *	use pointer from remove_uscore() as start of possible substitution characters
		 */
		for (p = convptr; *p; p++) {
			*subpos = *p;
			if (!is_duplicate(tempname,i)) {
				save_newname(t,tempname);
				names_hinted++;
				break;
			}
		}
		if (*p)
			continue;

		/*
		 *	last gasp: generate name
		 */
		for (c = '0'; c <= '9'; c++) {
			*subpos = c;
			if (!is_duplicate(tempname,i)) {
				save_newname(t,tempname);
				names_generated++;
				break;
			}
		}
		if (c > '9') {
			printf("\nWarning: can't generate newname for %s\n",t->origname);
			save_newname(t,t->origname);
			names_unchanged++;
		}
	}

	printf(" done!\n\n");

	/*
	 *	if undoing, we check after we've generated the names
	 */
	if (undo && !test) {
		printf("Checking old names ");
		for (i = 0, t = lfn; i < numfiles; i++, t++) {
			if (i%DOT_FREQUENCY == 0)
				printf(".");
			t->ext = 'c';
			sprintf(s,"%s\\%s.%c",dir,t->newname,t->ext);
			if (Fxattr(0,s,&xattr) < 0) {
				t->ext = 's';
				sprintf(s,"%s\\%s.%c",dir,t->newname,t->ext);
				if (Fxattr(0,s,&xattr) < 0) {
					printf("\nWarning: %s.o does not have a corresponding source file\n",t->newname);
					t->ext = NOT_FOUND;		/* source file does not exist */
					names_notfound++;
				}
			}
		}
		printf(" done!\n\n");
	}

	return 0;
}

int create_makefile(char *filename)
{
FILE *in, *out;
char outname[NAMELEN], bakname[NAMELEN];
int line;

	in = fopen(filename,"rt");
	if (!in) {
		printf("Can't open file %s\n",filename);
		return -1;
	}

	strmfe(outname,filename,NEW_EXT);
	strmfe(bakname,filename,BAK_EXT);
	out = fopen(outname,"wt");
	if (!out) {
		fclose(in);
		printf("Can't create file %s\n",outname);
		return -1;
	}

	for (line = 0; fgets(buf,BUFLEN,in); line++) {
		if (update_line(buf,line) < 0)
			return -1;
		if (fputs(buf,out) < 0) {
			printf("Can't write file %s\n",outname);
			return -1;
		}
	}

	fclose(in);
	fclose(out);

	/*
	 *	rename makefiles
	 */
	if (rename(filename,bakname) < 0) {
		printf("Can't rename %s => %s\n",filename,bakname);
		return -1;
	}
	if (rename(outname,filename) < 0) {
		printf("Can't rename %s => %s\n",outname,filename);
		return -1;
	}

	return 0;
}

int update_line(char *buf,int line)
{
static struct convert *t;
char *p, *q;

	if (line == 0) {	/* first time */
		t = lfn;
	}

	if (t->line > line)	/* no update needed */
		return 0;

	if (t->line < line) {
		printf("Programming error: t->line=%d, line=%d\n",t->line,line);
		return -1;
	}

	/*
	 *	this line needs updating: first we replace the tokens
	 */
	while (t->line == line)
		if (replace_token(buf,t++) < 0)
			return -1;
	/*
	 *	then we strip out the filler characters
	 */
	for (p = q = buf; *q; ) {
		if (*q == FILLER)
			q++;
		else *p++ = *q++;
	}
	*p = '\0';

	return 0;
}

int replace_token(char *buf,struct convert *t)
{
char *p = buf+t->offset;
char *q = t->newname;
int i;

	if (strncmp(p,t->origname,t->length) != 0) {
		printf("Filename %s.o not found in original position\n",t->origname);
		return -1;
	}

	for (i = 0; *q; i++)
		*p++ = *q++;

	for ( ; i < t->length; i++)
		*p++ = FILLER;

	return 0;
}

int rename_source(char *dir,int numfiles)
{
char oldname[NAMELEN], newname[NAMELEN];
struct convert *t;
int i;

	printf("Renaming source files ");
	if (test)
		printf("(simulation only)\n");
	for (i = 0, t = lfn; i < numfiles; i++, t++) {
		if ((i%DOT_FREQUENCY == 0) && !test)
			printf(".");
		if (strcmp(t->origname,t->newname) == 0)
			continue;
		sprintf(oldname,"%s\\%s.%c",dir,t->origname,t->ext);
		sprintf(newname,"%s\\%s.%c",dir,t->newname,t->ext);
		if (test)
			printf("%s => %s\n",oldname,newname);
		else if (rename(oldname,newname) < 0)
			printf("\nWarning: cannot rename %s to %s.%c\n",oldname,t->newname,t->ext);
	}

	printf(" done!\n\n");

	return 0;
}

int undo_changes(char *file,char *dir,int numfiles)
{
char oldname[NAMELEN], newname[NAMELEN];
struct convert *t;
int i;

	printf("Restoring original source file names ");
	if (test)
		printf("(simulation only)\n");
	for (i = 0, t = lfn; i < numfiles; i++, t++) {
		if ((i%DOT_FREQUENCY == 0) && !test)
			printf(".");
		if (strcmp(t->origname,t->newname) == 0)
			continue;
		sprintf(oldname,"%s\\%s.%c",dir,t->origname,t->ext);
		sprintf(newname,"%s\\%s.%c",dir,t->newname,t->ext);
		if (test)
			printf("%s => %s\n",newname,oldname);
		else if (rename(newname,oldname) < 0)
			printf("\nWarning: cannot restore %s to %s.%c\n",newname,t->origname,t->ext);
	}

	printf(" done!\n\n");

	strmfe(newname,file,NEW_EXT);
	strmfe(oldname,file,BAK_EXT);
	if (remove(newname) < 0) {
		printf("Warning: cannot remove temporary file %s\n",newname);
	}
	if (rename(file,newname) < 0) {
		printf("Warning: cannot rename current makefile %s to %s\n",file,newname);
	}
	if (rename(oldname,file) < 0) {
		printf("Warning: cannot rename original makefile %s to %s\n",oldname,file);
	}

	return 0;
}

/*
 *	removes underscores
 *
 *	returns pointer to start of possible (subsequent) substitution characters:
 *		a) if there are underscore(s) in the input following the last character
 *		   that would be used if the name were subsequently truncated, then
 *		   point to the last one, e.g.
 *				input:	v_delete_driver_info
 *				output:	vdeletedriverinfo
 *				returns ptr to i of info
 *		b) otherwise, point to character following the last character that would
 *		   be used if the name were subsequently truncated, e.g.
 *				input:	v_ftext_offset16
 *				output:	vftextoffset16
 *				returns ptr to second f of offset
 */
char *remove_uscore(char *new,char *old)
{
char *p, *q, *convptr = NULL;
int n = 0;

	for (p = new, q = old; *q; ) {
		switch(*q) {
		case '_':
			q++;
			if (n >= maxlen)		/* case (a) above */
				convptr = q;
			break;
		default:
			*p++ = *q++;
			if (++n == maxlen)		/* case (b) above */
				convptr = q;
		}
	}
	*p = '\0';

	if (!convptr)					/* paranoia */
		convptr = q;

	return convptr;
}

int create_listfile(char *filename,int numfiles)
{
FILE *out;
struct convert *t;
char outname[NAMELEN], temp[NAMELEN];
int i;

	printf("Creating list file\n");

	strmfe(outname,filename,LIST_EXT);
	out = fopen(outname,"wt");
	if (!out) {
		printf("Can't create list file %s\n",outname);
		return -1;
	}

	for (i = 0, t = lfn; i < numfiles; i++, t++) {
		if (t->ext == NOT_FOUND)	/* corresponding source does not exist */
			continue;
		sprintf(temp,"%s.o\n",t->newname);
		if (fputs(temp,out) < 0) {
			printf("Can't write file %s\n",outname);
			return -1;
		}
	}

	fclose(out);
	return 0;
}

